@@ -1,7 +1,7 @@
 import platform
 import subprocess
-from os import path, environ
-from typing import Optional, Callable, List, Coroutine, Any
+from os import path, environ, makedirs
+from typing import Optional, Coroutine
 
 from .connection import start_connection, CustomConnection
 from .objects import *
@@ -9,8 +9,6 @@ from .objects import *
 
 def get_proxy_name():
     machine = platform.machine().lower()
-    if "armv7" in machine:
-        return "proxy-armv7"
     if "arm64" in machine or "aarch64" in machine or "armv8" in machine:
         return "proxy-arm64"
     if "arm" in machine:
@@ -48,18 +46,17 @@ class IngeniumAPI:
     _connection: CustomConnection
     _proxy: subprocess.Popen
 
-    _detected_sock_fn: Optional[Callable[[int, str], None]] = None
+    _detected_sock_fn: Optional[Callable[[int, str], Coroutine[Any, Any, None]]] = None
     _detected_socks: List[int] = []
 
     _state = None
 
-    def __init__(self, state=None):
-        self._state = state
-        pass
+    _data_dir: Optional[str] = None
+    _onchange: Optional[Callable[[IngObject], None]] = None
 
-    def set_sock_detection(self, detected_fn: Callable[[int, str], Coroutine[Any, Any, None]], detected_socks: List[int]):
-        self._detected_sock_fn = detected_fn
-        self._detected_socks = detected_socks
+    def __init__(self, state=None):
+        self._state = state  # This state is only needed for 6lowpan
+        pass
 
     @property
     def state(self):
@@ -77,16 +74,54 @@ class IngeniumAPI:
         self._pass = None
         return self
 
-    async def load(self, just_login=False, hass_integration=False, debug=False):
+    def prepare_socks(self):
+        file_name = self.user if self.is_remote else self.host
+        file_path = self._data_dir + "/detected_socks_" + "".join(x for x in file_name if x.isalnum()) + ".dat"
+
+        makedirs(self._data_dir, exist_ok=True)
+
+        try:
+            with open(file_path) as f:
+                detected_socks = list(map(lambda x: int(x.split("#")[0]), f.readlines()))
+        except BaseException as e:
+            print("Error loading sock detect list", e)
+            detected_socks = []
+
+            try:
+                open(file_path, 'a').close()
+            except IOError:
+                pass
+
+        lock = asyncio.Lock()
+
+        async def sock_detected(addr: int, label: str):
+            async with lock:
+                if addr not in detected_socks:
+                    try:
+                        with open(file_path, 'a') as f2:
+                            f2.write(f"{addr}#{label}\n")
+                    except BaseException as e2:
+                        print("Error saving sock detect list", e2)
+                pass
+
+        self._detected_sock_fn = sock_detected
+        self._detected_socks = detected_socks
+
+    async def load(self, just_login=False, debug=False, data_dir=None, onchange: Callable[[IngObject], None] = None):
+        self._data_dir = data_dir
+        if data_dir is not None:
+            self.prepare_socks()
+
+        self._onchange = onchange
+
         my_env = environ.copy()
-        my_env["ENABLE_TRAFFICSERVER"] = str(self._user == "CalidadAire").lower()
         if debug:
             my_env["LOG_LEVEL"] = "debug"
         self._proxy = subprocess.Popen([PROXY_BIN], env=my_env, universal_newlines=True)
         await asyncio.sleep(1)
 
         try:
-            login_result = await start_connection(self, self._host, self._user, self._pass, just_login, hass_integration)
+            login_result = await start_connection(self, self._host, self._user, self._pass, just_login)
         except BaseException as e:
             print("Exception during login: " + repr(e) + " " + str(e))
             return False
@@ -97,10 +132,6 @@ class IngeniumAPI:
         # Add all the objects
         self._objects = []
         for o in login_result.get("devices"):
-            # Filter out any external devices
-            if o.get("externalId") is not None:
-                continue
-
             ctype = get_device_type(o.get("ctype"))
             address = safecast(o.get("address"), int)
 
@@ -109,32 +140,41 @@ class IngeniumAPI:
                 continue
 
             components = [IngComponent(c) for c in o.get("components")]
+            components_label = ", ".join([x.label for x in components])
 
-            if ctype.is_actuator():
-                self._objects.append(IngActuator(self, self._is_knx, address, ctype, components))
-            elif ctype.is_meterbus():
-                self._objects.append(IngMeterBus(self, self._is_knx, address, ctype, components))
-            elif ctype.is_tsif():
-                self._objects.append(IngSif(self, self._is_knx, address, ctype, components))
-            elif ctype.is_air_sensor():
-                self._objects.append(IngAirSensor(self, self._is_knx, address, ctype, components))
-            elif ctype.is_noise_sensor():
-                self._objects.append(IngNoiseSensor(self, self._is_knx, address, ctype, components))
-            elif ctype.is_busing_regulator():
-                self._objects.append(IngBusingRegulator(self, self._is_knx, address, ctype, components))
-            elif ctype.is_thermostat():
-                self._objects.append(IngThermostat(self, self._is_knx, address, ctype, components))
+            for component in components:
+                obj = None
+                if ctype.is_actuator():
+                    obj = IngActuator
+                elif ctype.is_meterbus():
+                    obj = IngMeterBus
+                elif ctype.is_tsif():
+                    obj = IngSif
+                elif ctype.is_air_sensor():
+                    obj = IngAirSensor
+                elif ctype.is_noise_sensor():
+                    obj = IngNoiseSensor
+                elif ctype.is_busing_regulator():
+                    obj = IngBusingRegulator
+                elif ctype.is_thermostat():
+                    obj = IngThermostat
+                elif ctype.is_three_phase_meter():
+                    obj = IngThreePhaseMeter
+                elif ctype.is_sensor_temp():
+                    obj = IngSensorTemp
+                if obj is not None:
+                    self._objects.append(obj(self, self._is_knx, address, ctype, component, components_label))
 
         if self.onload is not None:
-            self.onload()
+            self.onload(self)
         return True
 
     @property
-    def host(self) -> bool:
+    def host(self) -> str:
         return self._host
 
     @property
-    def user(self) -> bool:
+    def user(self) -> str:
         return self._user
 
     @property
@@ -168,8 +208,7 @@ class IngeniumAPI:
                 continue
 
             if obj.mode == ACTUATOR_NORMAL:
-                for comp in obj.components:
-                    entities.append((obj, comp))
+                entities.append(obj)
 
         return entities
 
@@ -180,8 +219,7 @@ class IngeniumAPI:
                 continue
 
             if obj.mode == ACTUATOR_BLIND:
-                for comp in obj.components:
-                    entities.append((obj, comp))
+                entities.append(obj)
 
         return entities
 
@@ -189,49 +227,57 @@ class IngeniumAPI:
         entities = []
         for obj in self.objects:
             if isinstance(obj, IngMeterBus):
-                for comp in obj.components:
-                    for i in range(1, 5):
-                        entities.append((obj, comp, i))
+                for i in range(1, 5):
+                    entities.append((obj, i))
         return entities
 
     def get_sifs(self):
         entities = []
         for obj in self.objects:
             if isinstance(obj, IngSif):
-                for comp in obj.components:
-                    for i in [0, 2, 3, 4]:
-                        entities.append((obj, comp, i))
+                for i in [0, 2, 3, 4]:
+                    entities.append((obj, i))
         return entities
 
     def get_air_sensors(self):
         entities = []
         for obj in self.objects:
             if isinstance(obj, IngAirSensor):
-                for comp in obj.components:
-                    entities.append((obj, comp, 0))
-                    entities.append((obj, comp, 1))
+                for i in [0, 1, 2, 3]:
+                    entities.append((obj, i))
         return entities
 
     def get_noise_sensors(self):
         entities = []
         for obj in self.objects:
             if isinstance(obj, IngNoiseSensor):
-                for comp in obj.components:
-                    entities.append((obj, comp))
+                entities.append(obj)
         return entities
 
     def get_lights(self):
         entities = []
         for obj in self.objects:
             if isinstance(obj, IngBusingRegulator):
-                for comp in obj.components:
-                    entities.append((obj, comp))
+                entities.append(obj)
         return entities
 
     def get_climates(self):
         entities = []
         for obj in self.objects:
             if isinstance(obj, IngThermostat):
-                for comp in obj.components:
-                    entities.append((obj, comp))
+                entities.append(obj)
+        return entities
+
+    def get_threephasemeters(self):
+        entities = []
+        for obj in self.objects:
+            if isinstance(obj, IngThreePhaseMeter):
+                entities.append(obj)
+        return entities
+
+    def get_sensor_temps(self):
+        entities = []
+        for obj in self.objects:
+            if isinstance(obj, IngSensorTemp):
+                entities.append(obj)
         return entities
