@@ -2,7 +2,7 @@ import asyncio
 import inspect
 import time
 from enum import Enum
-from typing import Callable, List, TYPE_CHECKING, Dict, Any
+from typing import Callable, List, TYPE_CHECKING, Dict, Any, Optional
 
 if TYPE_CHECKING:
     from .api import IngeniumAPI
@@ -20,18 +20,55 @@ class Package:
         return 'Package ( source={}, target={}, command={}, data1={}, data2={} )' \
             .format(self.source, self.target, self.command, self.data1, self.data2)
 
-    def as_bytes(self):
-        return bytes([
-            0xFF,
-            0xFF,
-            (self.target >> 8) & 0xFF,
-            self.target & 0xFF,
-            self.command & 0xFF,
-            self.data1 & 0xFF,
-            self.data2 & 0xFF,
-            0,
-            0  # To make it length 9
-        ])
+    @staticmethod
+    def from_bytes(b: bytes, offset: int, ident: Optional[bytes]) -> 'Package':
+        if ident is None:
+            return Package(
+                ((b[offset + 5] << 8) + b[offset + 6]) & 0xFFFF,
+                ((b[offset + 3] << 8) + b[offset + 4]) & 0xFFFF,
+                b[offset + 2],
+                b[offset + 7],
+                b[offset + 8]
+            )
+        else:
+            for i in range(8):
+                b[i + 2] ^= ident[i]
+            return Package(
+                ((b[offset + 4] << 8) + b[offset + 5]) & 0xFFFF,
+                ((b[offset + 2] << 8) + b[offset + 3]) & 0xFFFF,
+                b[offset + 6],
+                b[offset + 7],
+                b[offset + 8]
+            )
+
+    def as_bytes(self, ident: Optional[bytes]):
+        if ident is None:
+            return bytes([
+                0xFF,
+                0xFF,
+                (self.target >> 8) & 0xFF,
+                self.target & 0xFF,
+                self.command & 0xFF,
+                self.data1 & 0xFF,
+                self.data2 & 0xFF,
+                0,
+                0
+            ])
+        else:
+            b = bytes([
+                0xFF,
+                0xFF,
+                (self.target >> 8) & 0xFF,
+                self.target & 0xFF,
+                0xFE,
+                0xFE,
+                self.command & 0xFF,
+                self.data1 & 0xFF,
+                self.data2 & 0xFF
+            ])
+            for i in range(8):
+                b[i + 2] ^= ident[i]
+            return b
 
 
 class IngComponentType(Enum):
@@ -86,6 +123,9 @@ class IngComponentType(Enum):
 
     T_KNX_COLOR_PICKER = 9
 
+    T_THREE_PHASE_METER = 90  # Adeje
+    T_SENSOR_TEMP = 11  # Adeje
+
     def is_air_sensor(self): return self in [self.COD_AIR_QUALITY]
 
     def is_noise_sensor(self): return self in [self.COD_NOISE]
@@ -129,6 +169,10 @@ class IngComponentType(Enum):
 
     def is_busing_rgb_color_picker(self): return self in [self.COD_RGB]
 
+    def is_three_phase_meter(self): return self in [self.T_THREE_PHASE_METER]
+
+    def is_sensor_temp(self): return self in [self.T_SENSOR_TEMP]
+
 
 # noinspection PyBroadException
 def safecast(value, cast_fn):
@@ -152,33 +196,17 @@ class IngComponent:
     output: int
     icon: int
 
-    _notify_update: List[Callable]
-
     def __init__(self, json: Dict[str, Any] = None):
         if json is not None:
             self.id = json.get("id")
             self.label = json.get("label")
             self.output = safecast(json.get("output"), int)
             self.icon = safecast(json.get("icon"), int)
-        self._notify_update = []
 
     def __str__(self):
         return "Component ( id='{}', label='{}', output={} )" \
             .format(self.id, self.label, self.output)
 
-    def set_update_notify(self, callback: Callable):
-        self._notify_update = [callback]
-
-    def add_update_notify(self, callback: Callable):
-        self._notify_update.append(callback)
-
-    def update_notify(self):
-        for cb in self._notify_update:
-            try:
-                cb()
-            except BaseException as e:
-                print("Exception {} at {}".format(e, inspect.getsource(cb)))
-
 
 class IngObject:
     api: 'IngeniumAPI'
@@ -186,57 +214,62 @@ class IngObject:
     address: int
     type: IngComponentType
     type_name: str
-    components: List[IngComponent]
+
+    component: IngComponent
+    components_label: str
 
     available: bool
 
     def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
+                 comp: IngComponent, components_label: str):
         self.api = api
         self.is_knx = is_knx
         self.address = address
         self.type = ctype
         self.type_name = ctype.name.replace("COD_", "").replace("COD", "")
-        self.components = comps
+        self.component = comp
+        self.components_label = components_label
         self.available = False
 
     def __str__(self) -> str:
-        return "IngObject ( address='{}', type={}, components={} )" \
-            .format(self.address, self.type, [str(c) for c in self.components])
+        return "IngObject ( address='{}', type={}, component={} )" \
+            .format(self.address, self.type, str(self.component))
 
     async def update_state(self, p: Package) -> bool:
         raise NotImplementedError
 
     def get_info(self):
         return {
-            "name": ", ".join([x.label for x in self.components]),
+            "name": self.components_label,
             "identifiers": {("ingenium", self.address)},
             "model": self.type_name,
             "manufacturer": "Ingenium",
         }
 
+    def update_notify(self):
+        if self.api._onchange is not None:
+            try:
+                self.api._onchange(self)
+            except BaseException as e:
+                print("Exception on update_notify: {}".format(e, ))
+
 
 ACTUATOR_BLIND = 0
 ACTUATOR_NORMAL = 2
 
+
 class IngActuator(IngObject):
-    def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
-        super().__init__(api, is_knx, address, ctype, comps)
+    def __init__(self, *args):
+        super().__init__(*args)
 
-        self.blind1 = 0
-        self.blind2 = 0
-        self.blind3 = 0
-        self.blind4 = 0
-
-        self.updating_blinds = []
+        self.blind_value = 0
 
         self.consumption = -1
 
         self.current = -1
-        
+
         self.voltage = -1
-        self.voltages = []
+        self.voltage_ema = -1
         self.last_voltage = -1
         self.last_voltage_time = 0
 
@@ -249,15 +282,15 @@ class IngActuator(IngObject):
             self.is_sock = True
 
         self.output_state = 0
-        self.mode = ACTUATOR_BLIND if self.components[0].icon == 5 else ACTUATOR_NORMAL
+        self.mode = ACTUATOR_BLIND if self.component.icon == 5 else ACTUATOR_NORMAL
 
-    def get_switch_val(self, c: IngComponent):
-        power = 1 << c.output
+    def get_switch_val(self):
+        power = 1 << self.component.output
         return (self.output_state & power) == power
 
-    async def action_switch(self, c: IngComponent):
-        power = 1 << c.output
-        d2 = c.output + 8 if (self.output_state & power) == power else c.output
+    async def action_switch(self):
+        power = 1 << self.component.output
+        d2 = self.component.output + 8 if (self.output_state & power) == power else self.component.output
         p = Package(0xFFFF, self.address, 4, 2, d2)
         await self.api.send(p)
 
@@ -266,107 +299,34 @@ class IngActuator(IngObject):
         else:
             self.output_state &= ~(1 << (d2 - 8))
 
-        c.update_notify()
+        self.update_notify()
 
-    async def set_cover_val(self, c: IngComponent, value: int):
-        data1 = 0
-        if c.output < 2:
-            data1 = 4
-            self.blind1 = value
-        elif 2 <= c.output < 4:
-            data1 = 5
-            self.blind2 = value
-        elif 4 <= c.output < 6:
-            data1 = 6
-            self.blind3 = value
-        elif 6 <= c.output < 8:
-            data1 = 7
-            self.blind4 = value
-
-        p = Package(0xFFFF, self.address, 4, data1, value)
+    async def set_cover_val(self, value: int):
+        self.blind_value = value
+        p = Package(0xFFFF, self.address, 4, (self.component.output // 2) + 4, value)
         await self.api.send(p)
 
-        c.update_notify()
+        self.update_notify()
 
-    def get_cover_val(self, c: IngComponent):
-        if c.output < 2:
-            return self.blind1
-        elif 2 <= c.output < 4:
-            return self.blind2
-        elif 4 <= c.output < 6:
-            return self.blind3
-        elif 6 <= c.output < 8:
-            return self.blind4
-
-    async def read_state(self):
-        self.updating_blinds = []
-
-        for c in self.components:
-            if c.output < 2:
-                self.updating_blinds.append(4)
-            elif 2 <= c.output < 4:
-                self.updating_blinds.append(5)
-            elif 4 <= c.output < 6:
-                self.updating_blinds.append(6)
-            elif 6 <= c.output < 8:
-                self.updating_blinds.append(7)
-
-        await self._read()
-
-    async def _read(self):  # blind_number between 1 and 4
-        if len(self.updating_blinds) > 0:
-            val = self.updating_blinds[-1]
-
-            p = Package(0xFF, self.address, 3, val, 0)
-            await self.api.send(p)
+    def get_cover_val(self):
+        return self.blind_value
 
     async def maybe_detect_sock(self):
         if not self.is_sock:
             self.is_sock = True
             if self.api._detected_sock_fn is not None:
-                await self.api._detected_sock_fn(self.address, self.components[0].label)
+                await self.api._detected_sock_fn(self.address, self.component.label)
 
     async def update_state(self, p: Package) -> bool:
-        # This part is run when receiving the response to a manual read
-        if self.address == p.source and p.data1 == p.data2:
-            if len(self.updating_blinds) > 0:
-                val = self.updating_blinds.pop()
-                if val == 4:
-                    self.blind1 = p.data2
-                elif val == 5:
-                    self.blind2 = p.data2
-                elif val == 6:
-                    self.blind3 = p.data2
-                elif val == 7:
-                    self.blind4 = p.data2
-
-                if len(self.updating_blinds) > 0:
-                    await self._read()
-
-                return True
-
         if self.address != p.target:
             return False
 
         if self.mode == ACTUATOR_BLIND:
-            for c in self.components:
-                if c.output < 2 and p.data1 == 4:
-                    old_value = self.blind1
-                    self.blind1 = p.data2
-                    return old_value != self.blind1
-                elif 2 <= c.output < 4 and p.data1 == 5:
-                    old_value = self.blind2
-                    self.blind2 = p.data2
-                    return old_value != self.blind2
-                elif 4 <= c.output < 6 and p.data1 == 6:
-                    old_value = self.blind3
-                    self.blind3 = p.data2
-                    return old_value != self.blind3
-                elif 6 <= c.output < 8 and p.data1 == 7:
-                    old_value = self.blind4
-                    self.blind4 = p.data2
-                    return old_value != self.blind4
-                return False
+            if p.data1 == (self.component.output // 2) + 4:
+                old_value = self.blind_value
+                self.blind_value = p.data2
+                return old_value != self.blind_value
+            return False
 
         else:
             old_state = self.output_state
@@ -398,20 +358,22 @@ class IngActuator(IngObject):
             elif p.data1 == 87:
                 await self.maybe_detect_sock()
                 # Voltage
-                
-                self.voltages.append(p.data2)  
-                if len(self.voltages) > 5:     
-                    self.voltages.pop(0)
-
-                self.voltage = sum(self.voltages) // len(self.voltages)
+                # Exponential moving average
+                ema_alpha = 0.006
+                if self.voltage_ema < 0:
+                    self.voltage_ema = p.data2
+                self.voltage_ema = p.data2 * ema_alpha + self.voltage_ema * (1 - ema_alpha)
+                self.voltage = int(round(self.voltage_ema))
 
                 now = int(time.time())
-                if (now - self.last_voltage_time) > 60 and self.voltage != self.last_voltage:
+                time_diff = now - self.last_voltage_time
+                value_diff = abs(self.voltage - self.last_voltage)
+                if (time_diff > 100 and value_diff > 1) or value_diff > 3 or time_diff > 300:
                     self.last_voltage_time = now
                     self.last_voltage = self.voltage
                     return True
-     
                 return False
+
             elif p.data1 == 88:
                 await self.maybe_detect_sock()
                 old_value = self.active_power_bits[0]
@@ -438,80 +400,32 @@ METERBUS_FIELDS = {
 
 
 class IngMeterBus(IngObject):
-    def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
-        super().__init__(api, is_knx, address, ctype, comps)
-
-        self.updating_channels = []
+    def __init__(self, *args):
+        super().__init__(*args)
 
         self.factor = 1
 
-        self.available1: bool = False
-        self.available2: bool = False
-        self.available3: bool = False
-        self.available4: bool = False
-
-        self.cons1: int = 255
-        self.cons2: int = 255
-        self.cons3: int = 255
-        self.cons4: int = 255
-
-        self.thresh1: int = 255
-        self.thresh2: int = 255
-        self.thresh3: int = 255
-        self.thresh4: int = 255
+        self.chan_available = [False, False, False, False]
+        self.chan_value = [255, 255, 255, 255]
+        self.chan_threshold = [255, 255, 255, 255]
 
     # def action_set_limit(self, c: IngComponent):
     #    c.update_notify()
 
-    def get_cons(self, channel: int):
-        return [self.cons1, self.cons2, self.cons3, self.cons4][channel - 1]
+    def get_value(self, channel: int):
+        return self.chan_value[channel - 1]
 
     def get_thresh(self, channel: int):
-        return [self.thresh1, self.thresh2, self.thresh3, self.thresh4][channel - 1]
+        return self.chan_threshold[channel - 1]
 
     def get_available(self, channel: int):
-        return self.available and [self.available1, self.available2, self.available3, self.available4][channel - 1]
-
-    async def read_state(self):
-        self.updating_channels = list(METERBUS_FIELDS.keys())
-        await self._read()
-
-    async def _read(self):
-        if len(self.updating_channels) > 0:
-            chan = self.updating_channels[-1]
-
-            p = Package(0xFF, self.address, 3, chan, 0)
-            await self.api.send(p)
+        return self.available and self.chan_available[channel - 1]
 
     async def update_state(self, p: Package) -> bool:
-        if self.address == p.source:
-            if len(self.updating_channels) > 0:
-                mode = self.updating_channels.pop()
-
-                factor = 230 / 10
-                value = p.data2 * factor * self.factor
-                available = p.data2 != 255
-
-                if mode not in METERBUS_FIELDS:
-                    return False
-
-                key, available_key = METERBUS_FIELDS[mode]
-                old_val = getattr(self, key)
-                setattr(self, key, value)
-
-                if available_key is not None:
-                    setattr(self, available_key, available)
-
-                if len(self.updating_channels) > 0:
-                    await self._read()
-
-                return value != old_val
-
         if self.address != p.target:
             return False
 
-        if p.data1 == 9: # and p.data2 == 50:
+        if p.data1 == 9:  # and p.data2 == 50:
             self.factor = 2
             return False
 
@@ -519,17 +433,18 @@ class IngMeterBus(IngObject):
         value = p.data2 * factor * self.factor
         available = p.data2 != 255
 
-        if p.data1 not in METERBUS_FIELDS:
-            return False
+        if p.data1 in [0, 1, 2, 3]:
+            old_value = self.chan_threshold[p.data1]
+            self.chan_threshold[p.data1] = value
+            return old_value != value
 
-        key, available_key = METERBUS_FIELDS[p.data1]
-        old_val = getattr(self, key)
-        setattr(self, key, value)
+        if p.data1 in [10, 11, 12, 13]:
+            old_value = self.chan_value[p.data1 - 10]
+            self.chan_value[p.data1 - 10] = value
+            self.chan_available[p.data1 - 10] = available
+            return old_value != value
 
-        if available_key is not None:
-            setattr(self, available_key, available)
-
-        return value != old_val
+        return False
 
 
 SIF_AVAILABLE_MODES = [0, 2, 3, 4]
@@ -547,11 +462,8 @@ def convert_sif_value(value, mode: int):
 
 
 class IngSif(IngObject):
-    def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
-        super().__init__(api, is_knx, address, ctype, comps)
-
-        self.updating_modes = []
+    def __init__(self, *args):
+        super().__init__(*args)
 
         # [Temperature, Noise(unimplemented), Presence, Illuminance, Humidity]
         self.values = [0, 0, False, 0, 0]
@@ -560,32 +472,7 @@ class IngSif(IngObject):
 
         self.bat_baja = False
 
-    async def read_state(self):
-        self.updating_modes = SIF_AVAILABLE_MODES.copy()
-        await self._read()
-
-    async def _read(self):
-        if len(self.updating_modes) > 0:
-            val = self.updating_modes[-1]
-
-            p = Package(0xFF, self.address, 3, val, 0)
-            await self.api.send(p)
-
     async def update_state(self, p: Package) -> bool:
-        if self.address == p.source:
-            if len(self.updating_modes) > 0:
-                mode = self.updating_modes.pop()
-
-                if mode < self.values_len:
-                    old_value = self.values[mode]
-                    self.values[mode] = convert_sif_value(p.data2, mode)
-                    self.values_255[mode] = p.data2
-
-                    if len(self.updating_modes) > 0:
-                        await self._read()
-
-                    return self.values[mode] != old_value
-
         if self.address != p.target:
             return False
 
@@ -614,18 +501,20 @@ class IngSif(IngObject):
 
 
 class IngAirSensor(IngObject):
-    def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
-        super().__init__(api, is_knx, address, ctype, comps)
+    def __init__(self, *args):
+        super().__init__(*args)
 
-        self.bit_values = [0, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0]
-        self.values = [0, 0]
+        self.bit_values = [0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0]
+        self.values = [0, 0, 0, 0]
         self.thresh_values = [0, 0]
 
     async def update_state(self, p: Package) -> bool:
         if self.address != p.target:
             return False
 
+        if p.data1 > 13:
+            return False
+
         old_value = self.bit_values[p.data1]
         self.bit_values[p.data1] = p.data2
 
@@ -634,6 +523,16 @@ class IngAirSensor(IngObject):
             self.values[1] = (self.bit_values[2] << 8) | self.bit_values[3]
             return self.bit_values[3] != old_value
 
+        # Temperature
+        if p.data1 == 4 and self.bit_values[4] != 255:
+            self.values[2] = self.bit_values[4] / 2.0 - 25.0
+            return self.bit_values[4] != old_value
+
+        # Humidity
+        if p.data1 == 5 and self.bit_values[5] != 255:
+            self.values[3] = self.bit_values[5] / 2.0
+            return self.bit_values[5] != old_value
+
         if p.data1 == 13:
             self.thresh_values[0] = (self.bit_values[10] << 8) | self.bit_values[11]
             self.thresh_values[1] = (self.bit_values[12] << 8) | self.bit_values[13]
@@ -642,6 +541,9 @@ class IngAirSensor(IngObject):
         return False
 
     def get_available(self, mode: int):
+        if mode == 2 or mode == 3:
+            return self.values[2] != 0 or self.values[3] != 0
+
         # If the current mode value is max, or if _both_ mode values are 0, device is unavailable
         return self.values[mode] != 65535 and (self.values[0] != 0 or self.values[1] != 0)
 
@@ -656,9 +558,8 @@ class IngAirSensor(IngObject):
 
 
 class IngNoiseSensor(IngObject):
-    def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
-        super().__init__(api, is_knx, address, ctype, comps)
+    def __init__(self, *args):
+        super().__init__(*args)
 
         self.value = 255
         self.max = 255
@@ -697,9 +598,8 @@ class IngNoiseSensor(IngObject):
 
 
 class IngBusingRegulator(IngObject):
-    def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
-        super().__init__(api, is_knx, address, ctype, comps)
+    def __init__(self, *args):
+        super().__init__(*args)
 
         self.channels = [0, 0, 0, 0]
 
@@ -725,15 +625,19 @@ class IngBusingRegulator(IngObject):
 
 
 class IngThermostat(IngObject):
-    def __init__(self, api: 'IngeniumAPI', is_knx: bool, address: int, ctype: IngComponentType,
-                 comps: List[IngComponent]):
-        super().__init__(api, is_knx, address, ctype, comps)
+    def __init__(self, *args):
+        super().__init__(*args)
 
         self.temp = 0
         self.set_point = 0
         self.cmode = 0
         self.wmode = 0
 
+        self.humidity = None
+
+        self._co2 = NetworkNumber(2, 40, 1)
+        self._vocs = NetworkNumber(2, 42, 1)
+
     async def update_state(self, p: Package) -> bool:
         if self.address != p.target:
             return False
@@ -748,6 +652,11 @@ class IngThermostat(IngObject):
             self.set_point = p.data2 / 5
             return old != self.set_point
 
+        elif p.data1 == 2:
+            old = self.humidity
+            self.humidity = round(float(p.data2) * 100.0 / 255.0, 1)
+            return old != self.humidity
+
         elif p.data1 == 10:
             old = self.cmode
             self.cmode = p.data2
@@ -758,8 +667,19 @@ class IngThermostat(IngObject):
             self.wmode = p.data2
             return old != self.wmode
 
+        elif self._co2.update_state(p):
+            return True
+        elif self._vocs.update_state(p):
+            return True
+
         return False
 
+    def get_co2(self):
+        return self._co2.get_value()
+
+    def get_vocs(self):
+        return self._vocs.get_value()
+
     def get_mode(self):
         #  [HVAC_MODE_OFF, HVAC_MODE_HEAT, HVAC_MODE_COOL, HVAC_MODE_HEAT_COOL]
         if self.cmode == 0:
@@ -807,3 +727,181 @@ class IngThermostat(IngObject):
 
             p2 = Package(0xFFFF, self.address, 4, 14, mode + 4)
             await self.api.send(p2)
+
+
+class IngSensorTemp(IngObject):
+    def __init__(self, *args):
+        super().__init__(*args)
+        self.value = None
+
+    async def update_state(self, p: Package) -> bool:
+        if self.address != p.target:
+            return False
+
+        if p.data1 == self.component.output:
+            old_value = self.value
+            self.value = (float(p.data2) / 1.0) - 10.0
+            return old_value != self.value
+
+        return False
+
+
+class IngThreePhaseMeter(IngObject):
+    def __init__(self, *args):
+        super().__init__(*args)
+
+        self._current_power_factor = NetworkNumber(2, 180, 0)
+        self._frequency = NetworkNumber(2, 143, 0.01)
+
+        self._voltage = [NetworkNumber(2, 101 + i * 2, 0.1) for i in range(3)]
+        self._power_factor = [NetworkNumber(2, 125 + i * 2, 0.1) for i in range(3)]
+
+        self._current = [NetworkNumber(2, 107 + i * 2, 0) for i in range(3)]
+        self._active_power = [NetworkNumber(2, 113 + i * 2, 0) for i in range(3)]
+        self._active_energy = [NetworkNumber(4, 145 + i * 4, 0) for i in range(3)]
+        self._reactive_energy = NetworkNumber(4, 157, 0)
+
+        self._thdv = [NetworkNumber(2, 161 + i * 2, 0.01) for i in range(3)]
+        self._thdi = [NetworkNumber(2, 167 + i * 2, 0.01) for i in range(3)]
+
+    async def update_state(self, p: Package) -> bool:
+        if self.address != p.target:
+            return False
+
+        for i, n in enumerate(self._voltage):
+            if n.update_state(p):
+                return True
+        for i, n in enumerate(self._current):
+            if n.update_state(p):
+                return True
+        for i, n in enumerate(self._active_power):
+            if n.update_state(p):
+                return True
+        for i, n in enumerate(self._power_factor):
+            if n.update_state(p):
+                return True
+        if self._frequency.update_state(p):
+            return True
+        for i, n in enumerate(self._active_energy):
+            if n.update_state(p):
+                return True
+        if self._reactive_energy.update_state(p):
+            return True
+        for i, n in enumerate(self._thdv):
+            if n.update_state(p):
+                return True
+        for i, n in enumerate(self._thdi):
+            if n.update_state(p):
+                return True
+        if self._current_power_factor.update_state(p):
+            return True
+
+        return False
+
+    def phase_count(self):
+        return 3
+
+    def voltage(self, phase):
+        return self._voltage[phase].get_value()
+
+    def current(self, phase):
+        factor = self._current_power_factor.get_value_no_factor()
+        if factor is None:
+            factor = 30
+
+        value = self._current[phase].get_value_no_factor()
+        if value is None:
+            return None
+        return round(value * 10.0 / factor, 2)
+
+    def active_power(self, phase):
+        factor = self._current_power_factor.get_value_no_factor()
+        if factor is None:
+            factor = 30
+
+        value = self._active_power[phase].get_value_no_factor()
+        if value is None:
+            return None
+        return value * factor
+
+    def power_factor(self, phase):
+        return self._power_factor[phase].get_value()
+
+    def frequency(self):
+        return self._frequency.get_value()
+
+    def active_energy(self, phase):
+        factor = self._current_power_factor.get_value_no_factor()
+        if factor is None:
+            factor = 30
+
+        value = self._active_energy[phase].get_value_no_factor()
+        if value is None:
+            return None
+        return round(value * factor / 100.0, 2)
+
+    def reactive_energy(self):
+        factor = self._current_power_factor.get_value_no_factor()
+        if factor is None:
+            factor = 30
+
+        value = self._reactive_energy.get_value_no_factor()
+        if value is None:
+            return None
+        return round(value * factor / 100.0, 2)
+
+    def thdv(self, phase):
+        return self._thdv[phase].get_value()
+
+    def thdi(self, phase):
+        return self._thdi[phase].get_value()
+
+
+class NetworkNumber:
+    def __init__(self, size: int, address: int, factor: float):
+        assert size == 2 or size == 4 or size == 8
+
+        self._size = size
+        self._address = address
+        self._factor = factor
+        self._data = [0] * size
+        self._data_received = [False] * size
+
+        self._value = None
+
+    def get_value(self) -> Optional[float]:
+        if self._value is None:
+            return None
+
+        return float(self._value) * self._factor
+
+    def get_value_no_factor(self) -> Optional[float]:
+        return self._value
+
+    def update_state(self, p: Package) -> bool:
+        offset = p.data1 - self._address
+
+        if offset < 0 or offset >= self._size:
+            return False
+
+        self._data[offset] = p.data2
+        self._data_received[offset] = True
+
+        # Check that we received all the bytes, and return if not
+        for i in range(self._size):
+            if not self._data_received[i]:
+                return False
+
+        value = 0
+
+        # Reset the data received and build the value
+        for i in range(self._size):
+            self._data_received[i] = False
+            value <<= 8
+            value |= self._data[i] & 0xFF
+
+        if self._value == value:
+            return False
+
+        self._value = value
+        return True
